"""Module to visualize things in:

  - 1d (plot_1d.py), such as:

      - time-series
  - things in 2d (plot_2d.py), such as:

      - topoplot
      - time-frequency plots
      - mri images
  - things in 3d (plot_3d.py), such as:

      - surfaces (meshes)
      - channels
"""
from .plot_3d import Viz3
